Yusef.addPlugin
(
    'graphics',
    function Yusef_plugins_graphics()
    {
        var that            = this;
        var _config         = {};
        _config.pluginName  = 'Yusef.plugins.graphics';
        _config.cacheMaxAge = 1000*60*60*24*30;         //  30 days max age without being accessed
        _config.busyMaxAge  = 1000*60;                  //  1 minute between the busy notifications
        _config.busyString  = 'Generating preview images. Your computer may slow down.';



        var _lastTimeBusy   = 0;
        var _cache          = {};


        //  --
        var savedInfo = widget.preferenceForKey( _config.pluginName+'_cache' )||'';
        if( savedInfo!=='' )
        {
            _cache = JSON.parse( savedInfo );
        }
        PSO.pub( _config.pluginName+'_cache', _enumerateObject( _cache, '_cache RESTORED' ) );


        /*
         *  _updateCache
         */
        function _updateCache( imageFile )
        {
            PSO.pub( _config.pluginName+'_cache', imageFile, imageFile.exists );

            var touched     = false,
                timestamp   = new Date().getTime();

            if( imageFile && imageFile.exists )
            {
                //  set entry
                _cache[imageFile.path] =
                {
                    modified:   imageFile.modified.getTime(),
                    lastAccess: timestamp
                }
                touched = true;
                PSO.pub( _config.pluginName+'_cache', _enumerateObject( _cache, '_cache STORED' ) );
            }

            //  purge old entries
            timestamp   -= _config.cacheMaxAge;
            for( var k in _cache )
            {
                if( _cache[k].lastAccess<timestamp )
                {
                    touched     = true;
                    _cache[k]   = null;
                    delete _cache[k];
                }
            }

            if( touched )
            {
                //  save _originalImagesModified
                widget.setPreferenceForKey( JSON.stringify( _cache ), _config.pluginName+'_cache' );
            }
            PSO.pub( _config.pluginName+'_cache', _enumerateObject( _cache, '_cache STORED' ) );
        }

        /*
         *  _showBusyNotification
         */
        function _showBusyNotification( connection )
        {
            var now = new Date().getTime();
            if( connection /*&& connection.isOwner*/ && now-_config.busyMaxAge>_lastTimeBusy )
            {
                widget.showNotification( _config.busyString );
            }
            _lastTimeBusy   = now;
        }


        /*
         *  resizeImage
         */
        this.resizeImage = function( connection, imageFile, maxWidth, maxHeight, crop,  callback )
        {
            var image           = new Image(),
                crop            = crop||false,
                when            = new Date().getTime(),
                eventHandler    =
                (
                    //  eventHandler : bound & curried
                    function resizeImage_eventHandler( connection, maxWidth, maxHeight, crop, callback, when, event )
                    {
                        var eventType   = event.type.toLowerCase();

                        if( eventType=='error' )
                        {   //  error -> closeAndRedispatch for retry on the next request
                            connection.response.closeAndRedispatch();
                            return false;
                        }

                        //  load -> resize and callback
                        var srcImage    = event.srcElement,
                            canvas      = document.createElement( 'canvas' ),
                            tw          = srcImage.width,
                            th          = srcImage.height

                        if( crop )
                        {
                            //  crop resizing
                            var srcAR   = tw/th,
                                dstAR   = maxWidth/maxHeight;
                            
                            if( srcAR>dstAR )
                            {
                                if( th>maxHeight )
                                {
                                    tw *= maxHeight/th;
                                    th  = maxHeight;
                                }
                            }
                            else
                            {
                                if( tw>maxWidth )
                                {
                                    th *= maxWidth/tw;
                                    tw = maxWidth;
                                }
                            }

                            canvas.width    = Math.round( maxWidth )&-2;
                            canvas.height   = Math.round( maxHeight )&-2;
                            var context     = canvas.getContext( '2d' );
                            context.drawImage( srcImage,0,0, srcImage.width, srcImage.height, (maxWidth-tw)/2,(maxHeight-th)/2,tw,th );
                        }
                        else
                        {
                            //  fit resizing
                            if( tw>maxWidth )
                            {
                                th *= maxWidth/tw;
                                tw = maxWidth;
                            }
                            if( th>maxHeight )
                            {
                                tw *= maxHeight/th;
                                th = maxHeight;
                            }

                            canvas.width    = Math.round( tw )&-2;
                            canvas.height   = Math.round( th )&-2;
                            var context     = canvas.getContext( '2d' );
                            context.drawImage( srcImage,0,0, srcImage.width, srcImage.height, 0,0,tw,th );
                        }

                        context.globalCompositeOperation = 'destination-over';
                        context.fillStyle   = '#fff';
                        context.fillRect( 0,0, canvas.width, canvas.height );
                        

                        //  -> callback
                        callback( connection, canvas );

                        //  GC
                        delete  canvas;
                        delete  event.srcElement;
                        canvas  = null;
                        image   = null;
                    }
                ).bind(this).curry( connection, maxWidth, maxHeight, crop, callback, when );


            //  load -> resize image and call the callback with the canvas in argument
            image.addEventListener
            (
                'load',
                eventHandler,
                false
            );
            //  error -> callback with no argument
            image.addEventListener
            (
                'error',
                eventHandler,
                false
            );

            _showBusyNotification( connection );

            //  try to load the image
            image.src   = unescape( imageFile.path );
            return true;
        };


        /*
         *  serveResizedImage
         */
        this.serveResizedImage = function ( connection, imageFile, maxWidth, maxHeight, crop, quality )
        {
            connection.pending = true;

            var maxWidth    = maxWidth||64,
                maxHeight   = maxHeight||64,
                crop        = crop||false,
                quality     = quality||.5;

            connection.resizedImageFileName  = _hash( connection.request.uri+' in '+maxWidth+'x'+maxHeight+' @ '+quality +' '+ crop )+'.jpg';

            var fileObject  = Yusef.mountpoints.storage.resolve( connection.resizedImageFileName );
            var cached      = fileObject.exists===true && _cache[imageFile.path]!==undefined && _cache[imageFile.path].modified==imageFile.modified.getTime();


            //  update cache list
            _updateCache( imageFile );


            if( cached )
            {
                //  cached -> serve the resized image directly
                Yusef.serveFile( connection, fileObject );
            }
            else
            {
                //  !cached -> resize the image
                that.resizeImage
                (
                    connection,
                    imageFile,
                    maxWidth,
                    maxHeight,
                    crop,
                    function serveResizedImage_callback( connection, canvas )
                    {
                        var resizedImageFileStream = Yusef.mountpoints.storage.open( connection.resizedImageFileName, opera.io.filemode.WRITE );
                        if( resizedImageFileStream )
                        {
                            var dataUrl = canvas.toDataURL( 'image/jpeg', quality );
                            resizedImageFileStream.writeBase64( dataUrl.slice( dataUrl.indexOf(',')+1 ) );
                            resizedImageFileStream.close();

                            var fileObject  = Yusef.mountpoints.storage.resolve( connection.resizedImageFileName );
                            Yusef.serveFile( connection, fileObject );
                        }
                        else
                        {
                            Yusef.emitErrorPage( connection );
                        }
                    }
                );
            }
            return true;
        };
    }
);